@@ -51,6 +51,16 @@
 // TODO: this is turned on by default, but may not work for all architectures! Need to investigate.
 #endif
 
+// This allows others to supply their own random number generator function
+// which is assumed to be uniformly distributed over over 0 .. RAND_MAX
+inline long int randomDraw() {
+#ifndef __UNIFORM_RAND_OVERRIDE__
+  return rand();
+#else
+  return __UNIFORM_RAND_OVERRIDE__;
+#endif
+}
+
 using std::vector;
 using std::string;
 using std::pair;
@@ -61,45 +71,31 @@ template<typename T>
 struct Randomness {
   // Just a dummy class to avoid code repetition.
   // Owned by the AnnoyIndex, passed around to the distance metrics
-  Randomness() : _has_X2(true) {};
-  T _X1, _X2;
-  bool _has_X2;
-
-  inline T gaussian() {
-    // Taken from http://phoxis.org/2013/05/04/generating-random-numbers-from-normal-distribution-in-c/
-    _has_X2 = !_has_X2;
-    if (_has_X2)
-      return _X2;
-
-    T W, U1, U2;
-    do {
-      U1 = -1 + ((T)rand() / RAND_MAX) * 2;
-      U2 = -1 + ((T)rand() / RAND_MAX) * 2;
-      W = U1 * U1 + U2 * U2;
-    }
-    while (W >= 1 || W == 0);
-
-    T mult = sqrt((-2 * log(W)) / W);
-    _X1 = U1 * mult;
-    _X2 = U2 * mult;
-
-    return _X1;
-  }
-
   inline int flip() {
-    return rand() % 2;
+    return randomDraw() % 2;
   }
-  inline T uniform(T min, T max) {
-    return ((T)rand() / RAND_MAX) * (max - min) + min;
+  inline int index(int n) {
+    return randomDraw() % n;
+  }
+  inline void index2(int n, int* i, int* j) {
+    // Draw two random variables so that i != j
+    *i = index(n);
+    *j = index(n-1);
+    *j += (*j >= *i);
   }
 };
 
 template<typename T>
-inline void normalize(T* v, int f) {
+inline T get_norm(T* v, int f) {
   T sq_norm = 0;
   for (int z = 0; z < f; z++)
     sq_norm += v[z] * v[z];
-  T norm = sqrt(sq_norm);
+  return sqrt(sq_norm);
+}
+
+template<typename T>
+inline void normalize(T* v, int f) {
+  T norm = get_norm(v, f);
   for (int z = 0; z < f; z++)
     v[z] /= norm;
 }
@@ -131,10 +127,10 @@ struct Angular {
     // = a^2 / a^2 + b^2 / b^2 - 2ab/|a||b|
     // = 2 - 2cos
     T pp = 0, qq = 0, pq = 0;
-    for (int z = 0; z < f; z++) {
-      pp += x[z] * x[z];
-      qq += y[z] * y[z];
-      pq += x[z] * y[z];
+    for (int z = 0; z < f; z++, x++, y++) {
+      pp += (*x) * (*x);
+      qq += (*y) * (*y);
+      pq += (*x) * (*y);
     }
     T ppqq = pp * qq;
     if (ppqq > 0) return 2.0 - 2.0 * pq / sqrt(ppqq);
@@ -154,8 +150,14 @@ struct Angular {
       return random->flip();
   }
   static inline void create_split(const vector<node*>& nodes, int f, Randomness<T>* random, node* n) {
+    // Sample two random points from the set of nodes
+    // Calculate the hyperplane equidistant from them
+    int i, j;
+    random->index2(nodes.size(), &i, &j);
+    T i_norm = get_norm(nodes[i]->v, f);
+    T j_norm = get_norm(nodes[j]->v, f);
     for (int z = 0; z < f; z++)
-      n->v[z] = random->gaussian();
+      n->v[z] = nodes[i]->v[z] / i_norm - nodes[j]->v[z] / j_norm;
     normalize(n->v, f);
   }
 };
@@ -170,8 +172,8 @@ struct Euclidean {
   };
   static inline T distance(const T* x, const T* y, int f) {
     T d = 0.0;
-    for (int i = 0; i < f; i++) 
-      d += (x[i] - y[i]) * (x[i] - y[i]);
+    for (int i = 0; i < f; i++, x++, y++)
+      d += ((*x) - (*y)) * ((*x) - (*y));
     return d;
   }
   static inline T margin(const node* n, const T* y, int f) {
@@ -188,40 +190,37 @@ struct Euclidean {
       return random->flip();
   }
   static inline void create_split(const vector<node*>& nodes, int f, Randomness<T>* random, node* n) {
-    // See http://en.wikipedia.org/wiki/Bertrand_paradox_(probability)
-    // We want to sample a random hyperplane out of all hyperplanes that cut the convex hull.
-    // The probability of each angle is in proportion to the extent of the projection.
-    // This is good because it means we try to split in the longest direction.
-    // Doing this using Metropolis-Hastings sampling using 10 steps
-    T* v = (T*)malloc(sizeof(T) * f); // TODO: would be really nice to get rid of this allocation
-    double max_proj = 0.0;
-    for (int step = 0; step < 10; step++) {
-      for (int z = 0; z < f; z++)
-        v[z] = random->gaussian();
-      normalize(v, f);
-      // Project the nodes onto the vector and calculate max and min
-      T min = INFINITY, max = -INFINITY;
-      for (size_t i = 0; i < nodes.size(); i++) {
-        T dot = 0;
-        for (int z = 0; z < f; z++)
-          dot += nodes[i]->v[z] * v[z];
-        if (dot > max)
-          max = dot;
-        if (dot < min)
-          min = dot;
-      }
-      if (max - min > random->uniform(0, max_proj)) {
-        max_proj = max - min;
-        memcpy(n->v, v, sizeof(T) * f);
-        n->a = -random->uniform(min, max); // Take a random split along this axis
-      }
+    // Same as Angular version, but no normalization and has to compute the offset too
+    int i, j;
+    random->index2(nodes.size(), &i, &j);
+    n->a = 0.0;
+    for (int z = 0; z < f; z++) {
+      n->v[z] = nodes[i]->v[z] - nodes[j]->v[z];
+      n->a += -n->v[z] * (nodes[i]->v[z] + nodes[j]->v[z]) / 2;
     }
-    free(v);
   }
 };
 
+template<typename S, typename T>
+class AnnoyIndexInterface {
+ public:
+  virtual ~AnnoyIndexInterface() {};
+  virtual void add_item(S item, const T* w) = 0;
+  virtual void build(int q) = 0;
+  virtual bool save(const char* filename) = 0;
+  virtual void reinitialize() = 0;
+  virtual void unload() = 0;
+  virtual bool load(const char* filename) = 0;
+  virtual T get_distance(S i, S j) = 0;
+  virtual void get_nns_by_item(S item, size_t n, vector<S>* result, size_t search_k=-1) = 0;
+  virtual void get_nns_by_vector(const T* w, size_t n, vector<S>* result, size_t search_k=-1) = 0;
+  virtual S get_n_items() = 0;
+  virtual void verbose(bool v) = 0;
+  virtual void get_item(S item, vector<T>* v) = 0;
+};
+
 template<typename S, typename T, typename Distance>
-class AnnoyIndex {
+  class AnnoyIndex : public AnnoyIndexInterface<S, T> {
   /*
    * We use random projection to build a forest of binary trees of all items.
    * Basically just split the hyperspace into two sides by a hyperplane,
@@ -369,19 +368,19 @@ public:
     return true;
   }
 
-  inline T get_distance(S i, S j) {
+  T get_distance(S i, S j) {
     const T* x = _get(i)->v;
     const T* y = _get(j)->v;
     return Distance::distance(x, y, _f);
   }
 
-  void get_nns_by_item(S item, size_t n, vector<S>* result) {
+  void get_nns_by_item(S item, size_t n, vector<S>* result, size_t search_k=-1) {
     const typename Distance::node* m = _get(item);
-    _get_all_nns(m->v, n, result);
+    _get_all_nns(m->v, n, result, search_k);
   }
 
-  void get_nns_by_vector(const T* w, size_t n, vector<S>* result) {
-    _get_all_nns(w, n, result);
+  void get_nns_by_vector(const T* w, size_t n, vector<S>* result, size_t search_k=-1) {
+    _get_all_nns(w, n, result, search_k);
   }
   S get_n_items() {
     return _n_items;
@@ -390,6 +389,12 @@ public:
     _verbose = v;
   }
 
+  void get_item(S item, vector<T>* v) {
+    typename Distance::node* m = _get(item);
+    for (int z = 0; z < _f; z++)
+      v->push_back(m->v[z]);
+  }
+
 protected:
   void _allocate_size(S n) {
     if (n > _nodes_size) {
@@ -492,35 +497,20 @@ protected:
     return item;
   }
 
-  void _get_nns(const T* v, S i, vector<S>* result, S limit) {
-    const typename Distance::node* n = _get(i);
-
-    if (n->n_descendants == 0) {
-      // unknown item, nothing to do...
-    } else if (n->n_descendants == 1) {
-      result->push_back(i);
-    } else if (n->n_descendants <= _K) {
-      const S* dst = n->children;
-      result->insert(result->end(), n->children, &dst[n->descendants]);
-    } else {
-      bool side = Distance::side(n, v, _f, &_random);
-
-      _get_nns(v, n->children[side], result, limit);
-      if (result->size() < (size_t)limit)
-        _get_nns(v, n->children[!side], result, limit);
-    }
-  }
-
-  void _get_all_nns(const T* v, size_t n, vector<S>* result) {
+  void _get_all_nns(const T* v, size_t n, vector<S>* result, size_t search_k) {
     std::priority_queue<pair<T, S> > q;
 
+    if (search_k == -1)
+      search_k = n * _roots.size(); // slightly arbitrary default value
+
     for (size_t i = 0; i < _roots.size(); i++) {
       q.push(make_pair(numeric_limits<T>::infinity(), _roots[i]));
     }
 
     vector<S> nns;
-    while (nns.size() < n * _roots.size() && !q.empty()) {
+    while (nns.size() < search_k && !q.empty()) {
       const pair<T, S>& top = q.top();
+      T d = top.first;
       S i = top.second;
       const typename Distance::node* nd = _get(top.second);
       q.pop();
@@ -531,12 +521,14 @@ protected:
         nns.insert(nns.end(), nd->children, &dst[nd->n_descendants]);
       } else {
         T margin = Distance::margin(nd, v, _f);
-        q.push(make_pair(+margin, nd->children[1]));
-        q.push(make_pair(-margin, nd->children[0]));
+        q.push(make_pair(std::min(d, +margin), nd->children[1]));
+        q.push(make_pair(std::min(d, -margin), nd->children[0]));
       }
     }
 
-    std::sort(nns.begin(), nns.end());
+    // Get distances for all items
+    // To avoid calculating distance multiple times for any items, sort by id
+    sort(nns.begin(), nns.end());
     vector<pair<T, S> > nns_dist;
     S last = -1;
     for (size_t i = 0; i < nns.size(); i++) {
@@ -547,8 +539,10 @@ protected:
       nns_dist.push_back(make_pair(Distance::distance(v, _get(j)->v, _f), j));
     }
 
-    sort(nns_dist.begin(), nns_dist.end());
-    for (size_t i = 0; i < nns_dist.size() && result->size() < n; i++) {
+    size_t m = nns_dist.size();
+    size_t p = n < m ? n : m; // Return this many items
+    std::partial_sort(&nns_dist[0], &nns_dist[p], &nns_dist[m]);
+    for (size_t i = 0; i < p; i++) {
       result->push_back(nns_dist[i].second);
     }
   }
