@@ -19,78 +19,26 @@
 #include <stdint.h>
 
 
-template<typename S, typename T, typename Distance>
-class AnnoyIndexPython : public AnnoyIndex<S, T, Distance > {
-public:
+#if PY_MAJOR_VERSION >= 3
+#define IS_PY3K
+#endif
 
-  AnnoyIndexPython(int f): AnnoyIndex<S, T, Distance>(f) {}
+#ifndef Py_TYPE
+    #define Py_TYPE(ob) (((PyObject*)(ob))->ob_type)
+#endif
 
-  void add_item_py(S item, const vector<T> w) {
-    this->add_item(item, &w[0]);
-  }
-
-  PyObject* get_nns_by_item_py(S item, size_t n) {
-    PyObject* l = PyList_New(0);
-    vector<S> result;
-    this->get_nns_by_item(item, n, &result);
-    for (size_t i = 0; i < result.size(); i++) {
-      PyList_Append(l, PyInt_FromLong(result[i]));
-    }
-    return l;
-  }
-
-  PyObject* get_nns_by_vector_py(PyObject* v, size_t n) {
-    vector<T> w(this->_f);
-    for (int z = 0; z < PyList_Size(v) && z < this->_f; z++) {
-        PyObject *pf = PyList_GetItem(v,z);
-        w[z] = PyFloat_AsDouble(pf);
-    }
-    vector<S> result;
-    this->get_nns_by_vector(&w[0], n, &result);
-    PyObject* l = PyList_New(0);
-    for (size_t i = 0; i < result.size(); i++) {
-      PyList_Append(l, PyInt_FromLong(result[i]));
-    }
-    return l;
-  }
-
-  PyObject* get_item_vector_py(S item) {
-    const typename Distance::node* m = this->_get(item);
-    const T* v = m->v;
-    PyObject* l = PyList_New(0);
-    for (int z = 0; z < this->_f; z++) {
-      PyList_Append(l, PyInt_FromLong(v[z]));
-    }
-    return l;
-  }
-
-  bool save_py(const char* filename) {
-    return this->save(filename);
-  }
-
-  bool load_py(const char* filename) {
-    return this->load(filename);
-  }
-
-};
+#ifdef IS_PY3K
+    #define PyInt_FromLong PyLong_FromLong 
+#endif
 
 
-class AI : public AnnoyIndexPython <int32_t, float, Angular<int32_t, float> > {
-  public:
-  AI(int f) : AnnoyIndexPython <int32_t, float, Angular<int32_t, float> >(f){};
-};
-class EI : public AnnoyIndexPython <int32_t, float, Euclidean<int32_t, float> > {
-  public:
-  EI(int f) : AnnoyIndexPython <int32_t, float, Euclidean<int32_t, float> >(f){};
-};
-
+template class AnnoyIndexInterface<int32_t, float>;
 
 // annoy python object
 typedef struct {
   PyObject_HEAD
   int f;
-  char ch_metric;
-  void* ptr;
+  AnnoyIndexInterface<int32_t, float>* ptr;
 } py_annoy;
 
 
@@ -101,7 +49,6 @@ py_an_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
   self = (py_annoy *)type->tp_alloc(type, 0);
   if (self != NULL) {
     self->f = 0;
-    self->ch_metric = 0;
     self->ptr = NULL;
   }
 
@@ -115,16 +62,12 @@ py_an_init(py_annoy *self, PyObject *args, PyObject *kwds) {
 
   if (!PyArg_ParseTuple(args, "is", &self->f, &metric))
     return -1;
-  self->ch_metric = metric[0];
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
+  switch(metric[0]) {
   case 'a':
-    ai = new AI(self->f);
-    self->ptr = static_cast<AI*>(static_cast<void*>(ai));
+    self->ptr = new AnnoyIndex<int32_t, float, Angular<int32_t, float> >(self->f);
     break;
   case 'e':
-    ei = new EI(self->f);
-    self->ptr = static_cast<EI*>(static_cast<void*>(ei));
+    self->ptr = new AnnoyIndex<int32_t, float, Euclidean<int32_t, float> >(self->f);
     break;
   }
   return 0;
@@ -134,24 +77,15 @@ py_an_init(py_annoy *self, PyObject *args, PyObject *kwds) {
 static void 
 py_an_dealloc(py_annoy* self) {
   if (self->ptr) {
-    switch(self->ch_metric) {
-    case 'a':
-      delete (AI*)self->ptr;
-      break;
-    case 'e':
-      delete (EI*)self->ptr;
-      break;
-    }
+    delete self->ptr;
   }
-  self->ob_type->tp_free((PyObject*)self);
+  Py_TYPE(self)->tp_free((PyObject*)self);
 }
 
 
 static PyMemberDef py_annoy_members[] = {
   {(char*)"_f", T_INT, offsetof(py_annoy, f), 0,
    (char*)""},
-  {(char*)"ch_metric", T_CHAR, offsetof(py_annoy, ch_metric), 0,
-   (char*)"a|e for angular|euclidean"},
   {NULL}	/* Sentinel */
 };
 
@@ -164,22 +98,14 @@ py_an_load(py_annoy *self, PyObject *args) {
     return Py_None;
   if (!PyArg_ParseTuple(args, "s", &filename))
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    res = ai->load_py(filename);
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    res = ei->load_py(filename);
-    break;
-  }
+
+  res = self->ptr->load(filename);
+
   if (!res) {
     PyErr_SetFromErrno(PyExc_IOError);
     return NULL;
   }
-  return PyInt_FromLong(0);
+  Py_RETURN_TRUE;
 }
 
 
@@ -191,43 +117,30 @@ py_an_save(py_annoy *self, PyObject *args) {
     return Py_None;
   if (!PyArg_ParseTuple(args, "s", &filename))
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    res = ai->save_py(filename);
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    res = ei->save_py(filename);
-    break;
-  }
+
+  res = self->ptr->save(filename);
+
   if (!res) {
     PyErr_SetFromErrno(PyExc_IOError);
     return NULL;
   }
-  return PyInt_FromLong(0);
+  Py_RETURN_TRUE;
 }
 
 
 static PyObject* 
 py_an_get_nns_by_item(py_annoy *self, PyObject *args) {
-  PyObject* l = PyList_New(0);
-  int32_t item, n;
+  int32_t item, n, search_k=-1;
   if (!self->ptr) 
     return Py_None;
-  if (!PyArg_ParseTuple(args, "ii", &item, &n))
+  if (!PyArg_ParseTuple(args, "ii|i", &item, &n, &search_k))
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    l = ai->get_nns_by_item_py(item, (size_t)n);
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    l = ei->get_nns_by_item_py(item, (size_t)n);
-    break;
+
+  PyObject* l = PyList_New(0);
+  vector<int32_t> result;
+  self->ptr->get_nns_by_item(item, n, &result, search_k);
+  for (size_t i = 0; i < result.size(); i++) {
+    PyList_Append(l, PyInt_FromLong(result[i]));
   }
   return l;
 }
@@ -236,22 +149,22 @@ py_an_get_nns_by_item(py_annoy *self, PyObject *args) {
 static PyObject* 
 py_an_get_nns_by_vector(py_annoy *self, PyObject *args) {
   PyObject* v;
-  PyObject* l = PyList_New(0);
-  int32_t n;
+  int32_t n, search_k=-1;
   if (!self->ptr) 
     return Py_None;
-  if (!PyArg_ParseTuple(args, "Oi", &v, &n))
+  if (!PyArg_ParseTuple(args, "Oi|i", &v, &n, &search_k))
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    l = ai->get_nns_by_vector_py(v, (size_t)n);
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    l = ei->get_nns_by_vector_py(v, (size_t)n);
-    break;
+
+  vector<float> w(self->f);
+  for (int z = 0; z < PyList_Size(v) && z < self->f; z++) {
+    PyObject *pf = PyList_GetItem(v,z);
+    w[z] = PyFloat_AsDouble(pf);
+  }
+  vector<int32_t> result;
+  self->ptr->get_nns_by_vector(&w[0], n, &result, search_k);
+  PyObject* l = PyList_New(0);
+  for (size_t i = 0; i < result.size(); i++) {
+    PyList_Append(l, PyInt_FromLong(result[i]));
   }
   return l;
 }
@@ -259,25 +172,20 @@ py_an_get_nns_by_vector(py_annoy *self, PyObject *args) {
 
 static PyObject* 
 py_an_get_item_vector(py_annoy *self, PyObject *args) {
-  PyObject* l = NULL;
   int32_t item;
   if (!self->ptr) 
     return Py_None;
   if (!PyArg_ParseTuple(args, "i", &item))
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    l = ai->get_item_vector_py(item);
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    l = ei->get_item_vector_py(item);
-    break;
+
+  vector<float> v;
+  self->ptr->get_item(item, &v);
+  PyObject* l = PyList_New(0);
+  for (int z = 0; z < self->f; z++) {
+    PyList_Append(l, PyFloat_FromDouble(v[z]));
   }
-  if (l) return l;
-  return Py_None;
+
+  return l;
 }
 
 
@@ -285,7 +193,6 @@ static PyObject*
 py_an_add_item(py_annoy *self, PyObject *args) {
   vector<float> w;
 
-  //void add_item_py(S item, const vector<T> w) {
   PyObject* l;
   int32_t item;
   if (!self->ptr) 
@@ -296,18 +203,9 @@ py_an_add_item(py_annoy *self, PyObject *args) {
     PyObject *pf = PyList_GetItem(l,z);
     w.push_back(PyFloat_AsDouble(pf));
   }
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    ai->add_item_py(item, w);
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    ei->add_item_py(item, w);
-    break;
-  }
-  return PyInt_FromLong(0);
+  self->ptr->add_item(item, &w[0]);
+
+  Py_RETURN_NONE;
 }
 
 
@@ -318,18 +216,10 @@ py_an_build(py_annoy *self, PyObject *args) {
     return Py_None;
   if (!PyArg_ParseTuple(args, "i", &q))
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    ai->build(q);
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    ei->build(q);
-    break;
-  }
-  return PyInt_FromLong(0);
+
+  self->ptr->build(q);
+
+  Py_RETURN_TRUE;
 }
 
 
@@ -337,18 +227,10 @@ static PyObject *
 py_an_unload(py_annoy *self, PyObject *args) {
   if (!self->ptr) 
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    ai->unload();
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    ei->unload();
-    break;
-  }
-  return PyInt_FromLong(0);
+
+  self->ptr->unload();
+
+  Py_RETURN_TRUE;
 }
 
 
@@ -360,17 +242,9 @@ py_an_get_distance(py_annoy *self, PyObject *args) {
     return Py_None;
   if (!PyArg_ParseTuple(args, "ii", &i, &j))
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    d = ai->get_distance(i,j);
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    d = ei->get_distance(i,j);
-    break;
-  }
+
+  d = self->ptr->get_distance(i,j);
+
   return PyFloat_FromDouble(d);
 }
 
@@ -381,19 +255,10 @@ py_an_get_n_items(py_annoy *self, PyObject *args) {
   bool is_n=false;
   if (!self->ptr) 
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    n = ai->get_n_items();
-    is_n = true;
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    n = ei->get_n_items();
-    is_n = true;
-    break;
-  }
+
+  n = self->ptr->get_n_items();
+  is_n = true;
+
   if (is_n) return PyInt_FromLong(n);
   return Py_None;
 }
@@ -406,18 +271,10 @@ py_an_verbose(py_annoy *self, PyObject *args) {
     return Py_None;
   if (!PyArg_ParseTuple(args, "i", &verbose))
     return Py_None;
-  AI *ai; EI *ei;
-  switch(self->ch_metric) {
-  case 'a':
-    ai = static_cast<AI*>(self->ptr);
-    ai->verbose((bool)verbose);
-    break;
-  case 'e':
-    ei = static_cast<EI*>(self->ptr);
-    ei->verbose((bool)verbose);
-    break;
-  }
-  return PyInt_FromLong(0);
+
+  self->ptr->verbose((bool)verbose);
+
+  Py_RETURN_TRUE;
 }
 
 
@@ -438,8 +295,7 @@ static PyMethodDef AnnoyMethods[] = {
 
 
 static PyTypeObject PyAnnoyType = {
-  PyObject_HEAD_INIT(NULL)
-  0,                      /*ob_size*/
+  PyVarObject_HEAD_INIT(NULL, 0)
   "annoy.Annoy",          /*tp_name*/
   sizeof(py_annoy),       /*tp_basicsize*/
   0,                      /*tp_itemsize*/
@@ -483,21 +339,49 @@ static PyMethodDef module_methods[] = {
   {NULL}	/* Sentinel */
 };
 
-PyMODINIT_FUNC initannoylib(void) {
+#if PY_MAJOR_VERSION >= 3
+  static struct PyModuleDef moduledef = {
+    PyModuleDef_HEAD_INIT,
+    "annoylib",          /* m_name */
+    "",                  /* m_doc */
+    -1,                  /* m_size */
+    module_methods,      /* m_methods */
+    NULL,                /* m_reload */
+    NULL,                /* m_traverse */
+    NULL,                /* m_clear */
+    NULL,                /* m_free */
+  };
+#endif
+
+PyObject *create_module(void) {
   PyObject *m;
 
   if (PyType_Ready(&PyAnnoyType) < 0)
-    return;
+    return NULL;
 
+#if PY_MAJOR_VERSION >= 3
+  m = PyModule_Create(&moduledef);
+#else
   m = Py_InitModule("annoylib", module_methods);
-  if (m == NULL)
-    return;
+#endif
 
   if (m == NULL)
-    return;
+    return NULL;
 
   Py_INCREF(&PyAnnoyType);
   PyModule_AddObject(m, "Annoy", (PyObject *)&PyAnnoyType);
+  return m;
 }
 
+#if PY_MAJOR_VERSION >= 3
+  PyMODINIT_FUNC PyInit_annoylib(void) {
+    return create_module();      // it should return moudule object in py3
+  }
+#else
+  PyMODINIT_FUNC initannoylib(void) {
+    create_module();
+  }
+#endif
+
+
 // vim: tabstop=2 shiftwidth=2
